/*
 * dfu_func.c
 *
 * dfu_func provides the data structures, complex down- / upload functions
 * and dfu-suffix checking / handling functions
 *
 * Author: Jan Alte, DO1FJN
 *
 *
 * This file is part of the DFU Serial Loader (dfu-loader).
 * For general infomation about dfu-loader see "dfu_serial_main.c".
 *
 * dfu-loader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * dfu-loader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "dfu_func.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "checksums.h"
#include "serialIO.h"
#include "fileIO.h"



#define APP_DETACHWAITTIME	250000	// 250ms
#define CLRSTATUSWAITTIME	10000	// 10ms
#define DFU_MAXBLOCKSIZE	(SER_BUFSIZE-8)


typedef struct __attribute__((__packed__)) {
  union {
    tRS232pktHeader h;
    char data[8];    
  };
} tSERSmallAnswer;	// Paket reicht aus um 1 bis 3 Nutzbyte zu empfangen (ACK, NAK or Status)


typedef struct __attribute__((__packed__)) {
  union {
    tRS232pktHeader h;
    char data[24+6];    
  };
} tSERShortAnswer;	// Paket reicht aus um bis zu 24 Nutzbyte zu empfangen


typedef struct  __attribute__((__packed__)) {
  unsigned short Nr;
  unsigned short MaxSize;
} tDFUUploadReqBlock;


typedef struct  __attribute__((__packed__)) {
  union {
    struct {
      tRS232pktHeader Head;
      unsigned short  BlockNr;
      char	      BlockData[16];	//Dummy-Lenth
    };
    char	data[16];
  };
} tDFUUploadbuffer;

#define DFUUploadExtraLen	(sizeof(tRS232pktHeader)+2)


typedef struct  __attribute__((__packed__)) {
  union {
    struct {
      unsigned short  BlockNr;
      char	      BlockData[16];	//Dummy-Lenth
    };
    char	data[16];
  };
} tDFUDownloadbuffer;

#define DFUDownloadExtraLen	(2)



static const char DFUinvalStr[] = "INVALID";

static const char DFURequestStr[][16] = {
  "DFU_DETACH", "DFU_DOWNLOAD", "DFU_UPLOAD", "DFU_GETSTATUS",
  "DFU_CLRSTATUS", "DFU_GETSTATE", "DFU_ABORT"
};

static const char DFUStatusStr[][16] = {
  "OK", "errTARGET", "errFILE", "errWRITE",
  "errERASE", "errCHECK_ERASED", "errPROG", "errVERIFY",
  "errADDRESS", "errNOTDONE", "errFIRMWARE", "errVENDOR",
  "errUSBR", "errUNKNOWN", "errSTALLEDPKT"
};

static const char DFUStateStr[][16] = {
  "appIDLE", "appDETACH", "IDLE", "DNLOAD_SYNC",
  "DNLOAD_BUSY", "DOWNLOAD_IDLE", "MANIFEST_SYNC", "MANIFEST",
  "WAIT_RESET", "UPLOAD_IDLE", "ERROR"
};


const char *dfu_status_str(tDFUStatus Status) {
  if (Status <= errSTALLEDPKT) return DFUStatusStr[Status];
  return DFUinvalStr;
}

const char *dfu_state_str(tDFUDeviceState State) {
  if (State <= dfuERROR) return DFUStateStr[State];
  return DFUinvalStr;
}


void print_DFUinfo(const tDFUStatusInfo *Info) {
  printf("%s: %s\n", dfu_state_str(Info->State), dfu_status_str(Info->Status));
}

void bcd_version(unsigned short bcdVers, char* Result) {
  if (bcdVers&0xF000) {
    sprintf(Result, "%x%x.%x%x", (bcdVers>>12)&0xf, (bcdVers>>8)&0xf, (bcdVers>>4)&0xf, (bcdVers>>0)&0xf);
  } else {
    sprintf(Result, "%d.%x%x", (bcdVers>>8)&0xf, (bcdVers>>4)&0xf, (bcdVers>>0)&0xf);
  }
}


int dfu_append_filesuffix(int FILEfd, tDFUfilesuffix *Suffix) {
  Suffix->bcdDFU = 0x0100;		// DFU-Filesuffix Version 1.00
  Suffix->DFUsignature[0] = 'U';
  Suffix->DFUsignature[1] = 'F';
  Suffix->DFUsignature[2] = 'D';
  Suffix->length = (unsigned char)sizeof(tDFUfilesuffix);
  crc32Stream(&Suffix->crc, (char *)Suffix, sizeof(tDFUfilesuffix)-4);
  if (FILEfd>2) {
    int len;
    lseek(FILEfd, 0L, SEEK_END);
    len = write(FILEfd, Suffix, sizeof(tDFUfilesuffix));
    return (len != sizeof(tDFUfilesuffix))?-1:0;
  }
  return 0;
}


int dfu_read_filesuffix(int FILEfd, tDFUfilesuffix *Suffix, unsigned long *fwlength) {
  size_t filesize;
  char	 filebyte;
  int	 fileread;
  unsigned long filecrc = 0xffffffff;
  *fwlength = 0;
  if (FILEfd>2) {
    filesize = lseek(FILEfd, 0L, SEEK_END);
    if (filesize > sizeof(tDFUfilesuffix)) {
      lseek(FILEfd, 0L, SEEK_SET);	// calc crc of file...
      // crc32 of all byte must be calculatet as Zero...
      // but it's ok to compare crc(filelength-4) with the last 4 Bytes ("faster")
      for (fileread=filesize-4; fileread>0; fileread--) {
	read(FILEfd, &filebyte, 1);
	crc32(&filecrc, filebyte);
      } // rof
      lseek(FILEfd, -sizeof(tDFUfilesuffix), SEEK_END);
      read(FILEfd, Suffix, sizeof(tDFUfilesuffix));
      if (filecrc==Suffix->crc) {
	*fwlength = filesize - Suffix->length;
	return Suffix->length;	// returns Real Length of Suffix (maybe not readed)
      } else {
	return -sizeof(tDFUfilesuffix);
      }      
    } else return -1;	// File smaller than min-suffix
  }
  return 0;	// Ok, but Zerobytes written in Suffix
}


void dfu_show_suffixinfo(const tDFUfilesuffix *Suffix, int suffixlen, unsigned long fwlength) {
  char hVersionStr[6];
  switch (suffixlen) {
  case 0:
    printf("No filedescriptor to read suffix!\n");
    break;
  case -1:
    printf("File is smaller then minimum dfu-suffix.\n");
    break;
  default:
    if (suffixlen < -1) {	// invalid checksum
      suffixlen = -suffixlen;
      printf("Invalid checksum (%08Xh): File isn't a valid firmware or damaged.\n", Suffix->crc);
    } else {
      if (fwlength > 0x10000) {
        printf("Valid checksum (%08Xh): File is a valid firmware (%.1f KiB).\n", Suffix->crc, (float)fwlength/1024.0);
      } else {
        printf("Valid checksum (%08Xh): File is a valid firmware (%d Bytes).\n", Suffix->crc, fwlength);
      }
    }
    bcd_version(Suffix->bcdDFU, hVersionStr);
    printf("Device IDs %04X:%04X, loader release %X (bcd), DFU-header Version %s\n",
      Suffix->idVendor, Suffix->idProduct, Suffix->bcdDevice, hVersionStr);
    if (suffixlen > sizeof(tDFUfilesuffix)) 
      printf("Suffix has %d extra bytes.\n", suffixlen-sizeof(tDFUfilesuffix));
    break;
  } // hctiws
}


int dfu_test_filesuffix(const char *filename) {
  int FILEfd, suffixlen;
  unsigned long FWlen;
  tDFUfilesuffix Suffix;
  if (filename!=NULL) 
    FILEfd = open(filename, O_RDONLY);
  else
    return -1;
  if (FILEfd < 0) {
    file_error(errno, "OPEN FW FILE");
    return -1;
  }
  suffixlen = dfu_read_filesuffix(FILEfd, &Suffix, &FWlen);  
  dfu_show_suffixinfo(&Suffix, suffixlen, FWlen);
  close(FILEfd);
}



// *** Serielle Funktionen ***

int dfu_serial_getstatus(int SERfd, tDFUStatusInfo *Info) {
  int Len;
  tSERSmallAnswer Answer;
  Len = SER_SendReceive(SERfd, DFU_GETSTATUS, NULL, 0, Answer.data, sizeof(Answer));
  if (Len>0) {
    if (Len < 8) {
      fprintf(stderr, "%s: invalid packet received (to short)\n", DFURequestStr[DFU_GETSTATUS]);
      return -1;
    }
    if (Answer.h.cmd != (0x80|DFU_GETSTATUS)) {
      fprintf(stderr, "%s: invalid packet received (id)\n", DFURequestStr[DFU_GETSTATUS]);
      return -1;
    }
    memset(Info, 0,  sizeof(tDFUStatusInfo));	// Clear all information
    Info->Status = Answer.data[4];
    Info->State  = Answer.data[5];
    return 0;					// Success
  } else {
    fprintf(stderr, "%s: no answer received\n", DFURequestStr[DFU_GETSTATUS]);
  }
  return -1;
}


int dfu_serial_showstatus(int SERfd) {
  tDFUStatusInfo Status;	// To Get DFU_STATUS from device
  if (dfu_serial_getstatus(SERfd, &Status)==0) {
    printf("device is ");
    print_DFUinfo(&Status);
  }
}


int dfu_serial_getdevicedesc(int SERfd, tDFUdevicedescriptor *DeviceDescr) {
  int Len;
  tSERShortAnswer Answer;
  Len = SER_SendReceive(SERfd, DFUs_GETDEVICEDESCR, NULL, 0, Answer.data, sizeof(Answer));
  if (Len>0) {
    fflush(stdout);
    if (Len < (sizeof(tDFUdevicedescriptor)+6)) {
      fprintf(stderr, "GET_DEVDESC: invalid packet received (to short)\n");
      return -1;
    }
    if (Answer.h.cmd != (0x80|DFUs_GETDEVICEDESCR)) {
      fprintf(stderr, "GET_DEVDESC: invalid packet received (id)\n");
      return -1;
    }
    memcpy(DeviceDescr, Answer.data+4, sizeof(tDFUdevicedescriptor));	// Copy
    return sizeof(tDFUdevicedescriptor);	// Success    
  } else {
    fprintf(stderr, "GET_DEVDESC: no answer received\n");
  }
  return -1;
}


int dfu_serial_getfunctdesc(int SERfd, tDFUfunctionaldescriptor *FuncDescr) {
  int Len;
  tSERShortAnswer Answer;
  Len = SER_SendReceive(SERfd, DFUs_GETFUNCTIONDESCR, NULL, 0, Answer.data, sizeof(Answer));
  if (Len>0) {
    if (Len < (sizeof(tDFUfunctionaldescriptor)+6)) {
      fprintf(stderr, "GET_FUNCDESC: invalid packet received (to short)\n");
      return -1;
    }
    if (Answer.h.cmd != (0x80|DFUs_GETFUNCTIONDESCR)) {
      fprintf(stderr, "GET_FUNCDESC: invalid packet received (id)\n");
      return -1;
    }
    memcpy(FuncDescr, Answer.data+4, sizeof(tDFUfunctionaldescriptor));	// Copy
    return sizeof(tDFUfunctionaldescriptor);	// Success
  } else {
    fprintf(stderr, "GET_FUNCDESC: no answer received\n");
  }
  return -1;
}



int dfu_serial_settoidle(int SERfd) {
  int RetCode, LoopCnt;
  tDFUStatusInfo  Status;		// To Get DFU_STATUS from device
  tSERSmallAnswer Answer;
  LoopCnt = 3;
  do {
    RetCode = dfu_serial_getstatus(SERfd, &Status);
    if (RetCode < 0) return RetCode;
    printf("device is ");
    print_DFUinfo(&Status);
    switch(Status.State) {
      case appDETACH:
	usleep(APP_DETACHWAITTIME);
	break;
      case appIDLE:
	RetCode = SER_SendReceive(SERfd, DFU_DETACH, NULL, 0, Answer.data, sizeof(Answer));
	if (RetCode > 0) {
	  usleep(APP_DETACHWAITTIME);
	}
	break;
      case dfuIDLE:				// we are IDLE, do nothing except CLRSTATUS
	if (Status.Status == staOK) break;	// noError+IDLE -> Do nothing
      default:
	RetCode = SER_SendReceive(SERfd, DFU_CLRSTATUS, NULL, 0, Answer.data, sizeof(Answer));
	if (RetCode > 0) {
	  usleep(CLRSTATUSWAITTIME);
	}
	break;
    } // hctiws Status
    if (--LoopCnt == 0) {
      fprintf(stderr, "%s: Status '%s' can't be cleared to IDLE\n", DFURequestStr[DFU_GETSTATUS], dfu_status_str(Status.Status));
      return -1;
    } // fi Doof
  } while ((Status.State != dfuIDLE)||(Status.Status != staOK));
  return 0;
}



int dfu_serial_verify(int SERfd, const char *filename) {
  int FILEfd, RetCode, BlockCnt, BlockFailCnt;
  unsigned long FWlen, UpLoadSize;
  tDFUUploadReqBlock	   Block;	// 16Bit values: BlockNr & MaxBlockSize
  tDFUdevicedescriptor	   DeviceDescr;	// Devicedescriptor (ids & more)
  tDFUfunctionaldescriptor DeviceInfo;	// Basic Informations of DFU loader
  tDFUfilesuffix	   Suffix;	// to Append on written firmwarefile
  tDFUUploadbuffer	   *UpLoadBuf;	// Buffer for uploaded data (malloc)
  unsigned long		   UploadCrc;
  char			   *FileBuffer;
  unsigned long		   FileCrc;
  char			   LoaderVerStr[6];
  
  if (filename!=NULL) 
    FILEfd = open(filename, O_RDONLY);
  else
    return -1;
  if (FILEfd < 3) {
    file_error(errno, "VERIFY");
    return -1;
  }
  RetCode = dfu_read_filesuffix(FILEfd, &Suffix, &FWlen);
  dfu_show_suffixinfo(&Suffix, RetCode, FWlen);
  if (RetCode == -1) return -1;
  RetCode = dfu_serial_settoidle(SERfd);
  if (RetCode != 0) return RetCode;
  printf("Get Device Info... ");  
  RetCode = dfu_serial_getdevicedesc(SERfd, &DeviceDescr);
  if (RetCode < 0) return -1;
  RetCode = dfu_serial_getfunctdesc(SERfd, &DeviceInfo);
  if (RetCode < 0) return -1;
  
  Block.MaxSize = (DeviceInfo.TransferSize<DFU_MAXBLOCKSIZE)?DeviceInfo.TransferSize:DFU_MAXBLOCKSIZE;
  BlockCnt = (FWlen+Block.MaxSize-1)/Block.MaxSize;
  bcd_version(DeviceInfo.DFUVersion, LoaderVerStr);
  printf("\rDFU-Loader V%s (USB-ID %04X:%04X)  Blocksize %d Bytes (max. %d Bytes)\n",
    LoaderVerStr, DeviceDescr.idVendor, DeviceDescr.idProduct, Block.MaxSize, DeviceInfo.TransferSize);
  Block.Nr = 0;
  UpLoadSize = 0;	// Bytes of Received Data
  UpLoadBuf = (tDFUUploadbuffer *)malloc(Block.MaxSize+DFUUploadExtraLen); 
  if (UpLoadBuf == NULL) {
    printf("Error: Can't alloc uploadbuffer!\n");
    return -1;
  } // fi
  FileBuffer = (char *)malloc(Block.MaxSize);
  if (FileBuffer == NULL) {
    printf("Error: Can't alloc verify-buffer!\n");
    free(UpLoadBuf);
    return -1;
  } // fi
  lseek(FILEfd, 0L, SEEK_SET);
  BlockFailCnt = 0;
  FileCrc   = 0xffffffff;		// init Checksums
  UploadCrc = 0xffffffff;
  do {
    RetCode = SER_SendReceive(SERfd, DFU_UPLOAD, (char *)&Block, sizeof(Block), UpLoadBuf->data, Block.MaxSize+DFUUploadExtraLen);
    if (RetCode > 0) {		// Bytes received
      if (UpLoadBuf->Head.cmd != (0x80|DFU_UPLOAD)) {
	fprintf(stderr, "\nVERIFY: Error (id)\n");
	RetCode = -1;
      } // fi Err Answer
      if ((UpLoadBuf->Head.len > 3)&&(UpLoadBuf->BlockNr != Block.Nr)) {
	fprintf(stderr, "\nVERIFY: Error Blocknum (request %d, received %d)\n", Block.Nr, UpLoadBuf->BlockNr);
	RetCode = -1;	
      } // fi Err not the same Blocknumber
      if (RetCode > 0) {
	if (UpLoadBuf->Head.len > 3) {			// Upload Data received
	  int BlockLength = UpLoadBuf->Head.len-3;	// don't count Cmd, BlockNum
	  crc32Stream(&UploadCrc, UpLoadBuf->BlockData, BlockLength);
	  RetCode = read(FILEfd, FileBuffer, BlockLength);	// read Block from file (same size)
	  if (RetCode != BlockLength) {
	    if (RetCode < 0) {
	      printf("\nVERIFY: Error read from file '%s'.", filename);
	      file_error(errno, filename);
	    } else {
	      printf("\nVERIFY: EOF End of file reached before upload complete!\n");	      
	    }
            RetCode = -1;
	  } else {
	    crc32Stream(&FileCrc, FileBuffer, BlockLength);	    
	    if ((FileCrc != UploadCrc) || (memcmp(UpLoadBuf->BlockData, FileBuffer, BlockLength))) {
	      if (BlockFailCnt<10) {
		printf("\nBlock #%03d differs!\n", Block.Nr);
	      } else if (BlockFailCnt==10)
		printf("\nmore Blocks differs (not shown)...\n");
	      FileCrc = UploadCrc;	// Reset Check CRC for next Block
	      BlockFailCnt++;
	    } // fi CRC fails -> diff detects
	  }
          Block.Nr++;
          UpLoadSize += BlockLength;
	  if (UpLoadSize < 0x10000) {
	    printf("\rVerity Block #%03d/%03d (%d Bytes compared)...", Block.Nr, BlockCnt, UpLoadSize);
	  } else {
	    printf("\rVerify Block #%03d/%03d (%.1f KiB compared)...", Block.Nr, BlockCnt, (float)UpLoadSize/1024.0);	    
	  }
	} else {				// Upload finish
	  if (UpLoadBuf->data[4] == ACK) {
	    if (UpLoadSize==0) {
	      printf("\rNo Firmware found on device.\n");
	    } else {
	      printf("\rVerify Done. (%d Bytes in %d Blocks compared) CRC: %08lXh (w/o suffix)\n",
		UpLoadSize, Block.Nr, Suffix.crc);
            } // fi Err
	    RetCode = 0;
	  } else {
	    printf("\nVerify canceled by device after %d Bytes compared.\n", UpLoadSize);
	    if (UpLoadBuf->data[4] == NAK)
	      printf("VERIFY: Error NAK received. ");
	    RetCode = -1;	// Cancel
	  }
	} // esle FINISH
      } // fi RetCode>0
    } else {	// Error Receive
      printf("\nVERIFY: Error Receive (connection lost)\n");
    } // else Err RX
    fflush(stdout);	// Update Console Output
  } while (RetCode > 0);
  free(UpLoadBuf);
  free(FileBuffer);
  close(FILEfd);	// close Tempfile
  dfu_serial_showstatus(SERfd);
  if (BlockFailCnt == 0) {	// Verify success
    int WarnCnt = 0;
    if (FWlen != UpLoadSize) {
      if (FWlen < UpLoadSize)
	printf("Warning: %d Bytes readed from device (maybe DFU-filesuffix).\n", UpLoadSize-FWlen);
      else
	printf("Error: Firmwarefile contains %d Bytes more.\n", FWlen-UpLoadSize);
      WarnCnt++;
    } // fi Size differs
    if (Suffix.idVendor != DeviceDescr.idVendor) {
      printf("Warning: idVendor (device: %04X, filesuffix: %04X)\n", DeviceDescr.idVendor, Suffix.idVendor);
      WarnCnt++;
    }
    if (Suffix.idProduct != DeviceDescr.idProduct) {
      printf("Warning: idProduct (device: %04X, filesuffix: %04X)\n", DeviceDescr.idProduct, Suffix.idProduct);
      WarnCnt++;
    }
    if (Suffix.bcdDevice != DeviceDescr.bcdDevice) {
      printf("Warning: bcdDevice (device: %04X, filesuffix: %04X)\n", DeviceDescr.bcdDevice, Suffix.bcdDevice);
      WarnCnt++;
    }
    if (WarnCnt)
      printf("\nVERIFY result: Firmware is equal to file, but maybe not functional!\n");
    else
      printf("\nVERIFY SUCCESS.\n");
  } else {
    printf("VERIFY FAILED! %d Blocks are not equal to file\n", BlockFailCnt);
  }  
  fflush(stdout);	// Update Console Output
  return RetCode;
}


int dfu_serial_upload(int SERfd, const char *filename) {
  int FILEfd, RetCode;
  unsigned long UpLoadSize;
  char *tempfilename;
  tDFUUploadReqBlock	   Block;	// 16Bit values: BlockNr & MaxBlockSize
  tDFUdevicedescriptor	   DeviceDescr;	// Devicedescriptor (ids & more)
  tDFUfunctionaldescriptor DeviceInfo;	// Basic Informations of DFU loader
  tDFUfilesuffix	   Suffix;	// to Append on written firmwarefile
  tDFUUploadbuffer	   *UpLoadBuf;	// Buffer for uploaded data (malloc)
  char			   LoaderVerStr[6];
  
  RetCode = dfu_serial_settoidle(SERfd);
  if (RetCode != 0) return RetCode;
  printf("Get Device Info... ");  
  RetCode = dfu_serial_getdevicedesc(SERfd, &DeviceDescr);
  if (RetCode < 0) return -1;
  Suffix.bcdDevice = DeviceDescr.bcdDevice;	// Device (Loader) Release No.
  Suffix.idVendor  = DeviceDescr.idVendor;
  Suffix.idProduct = DeviceDescr.idProduct;
  RetCode = dfu_serial_getfunctdesc(SERfd, &DeviceInfo);
  if (RetCode < 0) return -1;
  Block.MaxSize = (DeviceInfo.TransferSize<DFU_MAXBLOCKSIZE)?DeviceInfo.TransferSize:DFU_MAXBLOCKSIZE;
  bcd_version(DeviceInfo.DFUVersion, LoaderVerStr);
  printf("\rDFU-Loader V%s (USB-ID %04X:%04X)  Blocksize %d Bytes (max. %d Bytes)\n",
    LoaderVerStr, DeviceDescr.idVendor, DeviceDescr.idProduct, Block.MaxSize, DeviceInfo.TransferSize);
  Block.Nr = 0;
  UpLoadSize = 0;	// Bytes of Received Data
  UpLoadBuf = (tDFUUploadbuffer *)malloc(Block.MaxSize+DFUUploadExtraLen);
  if (UpLoadBuf == NULL) {
    printf("Error: Can't alloc uploadbuffer!\n");
    return -1;
  } // fi
  // open temporary firmware file
  tempfilename = NULL;
  if (filename != NULL) {
    tempfilename = (char *)malloc(strlen(filename)+5);
    if (tempfilename!=NULL) sprintf(tempfilename, "%s.tmp", filename);
  }  
  if (tempfilename!=NULL) 
    FILEfd = open(tempfilename, O_WRONLY|O_CREAT|O_TRUNC|O_APPEND, 0644);
  else
    FILEfd = -1;
  Suffix.crc = 0xffffffff;	// init Checksum
  do {
    RetCode = SER_SendReceive(SERfd, DFU_UPLOAD, (char *)&Block, sizeof(Block), UpLoadBuf->data, Block.MaxSize+DFUUploadExtraLen);
    if (RetCode > 0) {		// Bytes received
      if (UpLoadBuf->Head.cmd != (0x80|DFU_UPLOAD)) {
	fprintf(stderr, "\nUPLOAD: Error (id)\n");
	RetCode = -1;
      } // fi Err Answer
      if ((UpLoadBuf->Head.len > 3)&&(UpLoadBuf->BlockNr != Block.Nr)) {
	fprintf(stderr, "\nUPLOAD: Error Blocknum (request %d, received %d)\n", Block.Nr, UpLoadBuf->BlockNr);
	RetCode = -1;	
      } // fi Err not the same Blocknumber
      if (RetCode > 0) {
	if (UpLoadBuf->Head.len > 3) {			// Upload Data received
	  int BlockLength = UpLoadBuf->Head.len-3;	// don't count Cmd, BlockNum
	  crc32Stream(&Suffix.crc, UpLoadBuf->BlockData, BlockLength);
	  if (FILEfd > 2) {	// write only if valid FILEfd
	    RetCode = write(FILEfd, UpLoadBuf->BlockData, BlockLength);
	    if (RetCode != BlockLength) {	// Error couldn't write hole block
	      printf("\nUPLOAD: Error write to file '%s'.", tempfilename);
	      file_error(errno, tempfilename);
	      RetCode = -1;
	    } // fi Err
	  } // fi write in real files/devices only (not in stdout/err)
          Block.Nr++;
          UpLoadSize += BlockLength;
	  if (UpLoadSize < 0x10000) {
	    printf("\rUploading Block #%03d (%d Bytes received)...", Block.Nr, UpLoadSize);
	  } else {
	    printf("\rUploading Block #%03d (%.1f KiB received)...", Block.Nr, (float)UpLoadSize/1024.0);	    
	  }
	} else {				// Upload finish
	  if (UpLoadBuf->data[4] == ACK) {
	    if (UpLoadSize==0) {
	      printf("\rNo Firmware found on device (no file written).\n");
	      RetCode = 0;
	    } else {
	      RetCode = dfu_append_filesuffix(FILEfd, &Suffix);	// Append DFU-Filesuffix
	      printf("\rUpload Done. (%d Bytes in %d Blocks received) CRC: %08lXh (w/o suffix)\n", UpLoadSize, Block.Nr, Suffix.crc);
	      if (RetCode!=0) {
                printf("\nSUFFIX: Error append suffix to file.", tempfilename);
                file_error(errno, tempfilename);
	      }
            } // fi Err
	  } else {
	    printf("\nUpload canceled by device after %d Bytes received.\n", UpLoadSize);
	    if (UpLoadBuf->data[4] == NAK)
	      printf("UPLOAD: Error NAK received. ");
	    RetCode = -1;	// Cancel
	  }
	} // esle FINISH
      } // fi RetCode>0
    } else {	// Error Receive
      printf("\nUPLOAD: Error Receive (connection lost)\n");
    } // else Err RX
    fflush(stdout);	// Update Console Output
  } while (RetCode > 0);
  if (FILEfd > 2) {
    close(FILEfd);	// close Tempfile
    if (UpLoadSize > 0)
      file_rename(tempfilename, filename);
    else
      file_delete(tempfilename);
  } // fi have file
  if (tempfilename!=NULL) free(tempfilename);
  free(UpLoadBuf);
  dfu_serial_showstatus(SERfd);
  fflush(stdout);	// Update Console Output
  return RetCode;
}



#define POLLSTATUSINTERVAL	10000	// 10ms
#define POLLSTATUSTIMEOUT	(2000000L/POLLSTATUSINTERVAL)

int dfu_serial_pollstatus(int SERfd, tDFUDeviceState expectedState) {
  int RetCode, LoopCnt;
  tDFUStatusInfo  Status;		// To Get DFU_STATUS from device
  tSERSmallAnswer Answer;
  LoopCnt = POLLSTATUSTIMEOUT;
  do {
    usleep(POLLSTATUSINTERVAL);
    RetCode = dfu_serial_getstatus(SERfd, &Status);
    if (RetCode < 0) return RetCode;
//    print_DFUinfo(&Status);
    if (Status.State == dfuERROR) {
      fprintf(stderr, "\nDOWNLOAD: State is '%s' (Status: '%s').\n", 
	dfu_state_str(Status.State), dfu_status_str(Status.Status));
      return -1;	// Error!
    } else if (--LoopCnt == 0) {
      fprintf(stderr, "\nDOWNLOAD: State stalled in '%s'.\n", dfu_state_str(Status.State));
      return -1;
    } // fi Doof    
  } while ((Status.State!=dfuIDLE) && (Status.State!=expectedState));  
  return 0; 
}


int dfu_serial_download(int SERfd, const char *filename) {
  int FILEfd, RetCode, SuffixLen;
  int BlockSize, readedSize;
  unsigned long FWlen, DownLoadSize;
  tDFUfilesuffix	   Suffix;	// Filesuffix (validate)
  tDFUdevicedescriptor	   DeviceDescr;	// Devicedescriptor (ids & more)
  tDFUfunctionaldescriptor DeviceInfo;	// Basic Informations of DFU loader
  tSERSmallAnswer	   Answer;
  tDFUDownloadbuffer	   *DnLoadBuf;	// Buffer for download (malloc)
  char			   LoaderVerStr[6];
  
  if (filename!=NULL) 
    FILEfd = open(filename, O_RDONLY);
  else
    return -1;
  if (FILEfd < 3) {
    file_error(errno, "DOWNLOAD");
    return -1;
  } 
  SuffixLen = dfu_read_filesuffix(FILEfd, &Suffix, &FWlen);
  dfu_show_suffixinfo(&Suffix, SuffixLen, FWlen);
  RetCode = dfu_serial_settoidle(SERfd);
  if (RetCode != 0) return RetCode;
  printf("Get Device Info... ");  
  RetCode = dfu_serial_getdevicedesc(SERfd, &DeviceDescr);
  if (RetCode < 0) return -1;
  RetCode = 0;
  if ((Suffix.idVendor != 0xFFFF)&&(Suffix.idVendor!=DeviceDescr.idVendor)) {
    RetCode = 1;
  } // fi Check idVendor
  if ((Suffix.idProduct != 0xFFFF)&&(Suffix.idProduct!=DeviceDescr.idProduct)) {
    RetCode |= 2;
  } // fi Check idProduct
  switch (RetCode) {
    case 0:
      break;
    case 1:
      fprintf(stderr, "idVendor doesn't match (device %04X, firmwarefile %04X)!\n",  DeviceDescr.idVendor,  Suffix.idVendor);
      return -1;
    case 2:
      fprintf(stderr, "idProduct doesn't match (device %04X, firmwarefile %04X)!\n", DeviceDescr.idProduct, Suffix.idProduct);
      return -1;
    case 3:
      fprintf(stderr, "idVendor and idProduct doesn't match (device %04X:%04X, firmwarefile %04X:%04X)!\n", 
        DeviceDescr.idVendor, DeviceDescr.idProduct, Suffix.idVendor, Suffix.idProduct);
      return -1;
    default:
      fprintf(stderr, "Unknown Error %d while checking firmware.\n");
      return -2;
  } // hctiws  
  RetCode = dfu_serial_getfunctdesc(SERfd, &DeviceInfo);
  if (RetCode < 0) return -1;
  BlockSize = (DeviceInfo.TransferSize<DFU_MAXBLOCKSIZE)?DeviceInfo.TransferSize:DFU_MAXBLOCKSIZE;
  DnLoadBuf = (tDFUDownloadbuffer *)malloc(BlockSize+DFUDownloadExtraLen);
  if (DnLoadBuf == NULL) {
    printf("Error: Can't alloc downloadbuffer!\n");
    return -1;
  } // fi
  bcd_version(DeviceInfo.DFUVersion, LoaderVerStr);
  printf("\rDFU-Loader V%s (USB-ID %04X:%04X)  Used Blocksize for download: %d Bytes\n",
    LoaderVerStr, DeviceDescr.idVendor, DeviceDescr.idProduct, BlockSize);
  // Start Downloading
  DnLoadBuf->BlockNr = 0;
  DownLoadSize = 0;
  lseek(FILEfd, 0L, SEEK_SET);	// Set File to Begin
  do {
    if (DownLoadSize >= FWlen) {		// complete?
      readedSize = 0;	// Last Packet to enter dfuMANIFESTATION
    } else {					// NOT complete.
      if (BlockSize > (FWlen-DownLoadSize)) {	// reduce last Block
        BlockSize = (FWlen-DownLoadSize);
      } // fi Last
      readedSize = read(FILEfd, DnLoadBuf->BlockData, BlockSize);
      if (readedSize != BlockSize) {
	if (readedSize < 0)
          file_error(errno, "DOWNLOAD");
        else
	  printf("DOWNLOAD: unexpected end of file (%d Bytes read, %d expected)\n", readedSize, BlockSize);
	readedSize = -1;
      } // fi file error
    } // esle last
    
    if ((readedSize >= 0) && (dfu_serial_pollstatus(SERfd, dfuDNLOAD_IDLE)==0)) {
      RetCode = SER_SendReceive(SERfd, DFU_DNLOAD, DnLoadBuf->data, readedSize+DFUDownloadExtraLen, Answer.data, sizeof(Answer));   
    } else {
      RetCode = -1;
    }
    if (RetCode > 0) {		// Bytes received 
      if (Answer.h.cmd != (0x80|DFU_DNLOAD)) {
	fprintf(stderr, "\nDOWNLOAD: Error (id)\n");
	RetCode = -1;
      } // fi Err Answer
      if (RetCode == 7) {
	switch (Answer.data[4]) {
	  case ACK:	// Downloaded BLOCK was accepted. Now poll Status...
	    if (readedSize==0) {	// last Block - we are finish!
	      printf("\nFinish downloading. Wait for manifestation...");
	      RetCode = dfu_serial_pollstatus(SERfd, dfuMANIFEST_WAIT_RESET);
	      if (RetCode==0) {
	        printf("\rDownload finished.                           \n");
	      } // fi
	    } else {
	      DownLoadSize += readedSize;
	      DnLoadBuf->BlockNr++;	// next Block
	      if ((DownLoadSize < 0x10000)&&(FWlen<0x10000)) {
		printf("\rProgrammimg Block #%03d (%d Bytes, %d total)...", DnLoadBuf->BlockNr, DownLoadSize, FWlen);
	      } else {
		printf("\rProgramming Block #%03d (%.1f KiB of %.1f KiB total)...", DnLoadBuf->BlockNr,
		  (float)DownLoadSize/1024.0, (float)FWlen/1024.0);	    
	      }
	      fflush(stdout);	// Update Console Output	      
	    } // esle not finish
	    break;
	  case NAK:
	    fprintf(stderr, "\nDOWNLOAD: NAK packet received.\n");
	    RetCode = -1;
	    break;
	  default:
	    fprintf(stderr, "\nDOWNLOAD: Invalid paket received (ACK/NAK expected, %02Xh received).\n", Answer.data[4]);
	    RetCode = -1;
	    break;
	} // hctiws
      } else {
	fprintf(stderr, "\nDOWNLOAD: ACK/NAK paket expected.\n");
	RetCode = -1;
      } // esle not a 7-Byte answer  
    } else {	// Error Receive
      printf("\nDOWNLOAD: Error Receive (connection lost).\n");
    } // else Err RX
    fflush(stdout);	// Update Console Output
  } while ((readedSize > 0) && (RetCode >= 0));  
  free(DnLoadBuf);
  close(FILEfd);
  dfu_serial_showstatus(SERfd);
  fflush(stdout);	// Update Console Output
  return RetCode;
}

