/*
 * serialIO.c
 *
 * simple device file based serial functions
 *
 * Author: Jan Alte, DO1FJN
 *
 *
 * This file is part of the DFU Serial Loader (dfu-loader).
 * For general infomation about dfu-loader see "dfu_serial_main.c".
 *
 * dfu-loader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * dfu-loader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
Report:
2005-12-05  JA	Serieller Filedescriptor SERfd als Parameter
2006-06-13  JA	Fehler SER_Sendblock bei ARM9 (!= BigEndian)

2009-08-27  JA	Version für Linux und größeren Paketen für dfu-loader Projekt
2009-08-29  JA	FrameStartID ist D0h für Extra-Long Packets

*/

#include "serialIO.h"
#include "checksums.h"
#include "fileIO.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <time.h>
#include <errno.h>

#define swap16(val)	(((val>>8)&0x00FF)|((val<<8)&0xFF00))


struct termios OldTermIOS;

tRS232buffer SERbuffer;
unsigned int SERrxlen;
clock_t  last_rx_time;

const char SER_DefaultDevice[] = SER_DEVICE;

int SER_open(const char *devicefile) {
  struct termios termios;
  int SERfd = 0;
  // Open port, no wait/no block for reading chars
  if (devicefile == NULL) devicefile = SER_DefaultDevice;
  SERfd = open(devicefile, O_RDWR | O_NONBLOCK | O_NOCTTY | O_NDELAY);
  if (SERfd == -1) {
    SERfd = 0;
    file_error(errno, devicefile);
    logError(0, " cant'b be open.");
    exit(-1);
  } /* endif open */
  fcntl(SERfd, F_SETFL, 0);
  // Setup speed, Line- and Bit-parameters
  // get old attribute
  if (tcgetattr(SERfd, &termios)) {
    // This is not a serial device
    logError(errno, "tcgetattr() Can't read device configuration. Is this a serial device?");
    SER_close(SERfd);
    exit(-1);
  }
  OldTermIOS = termios;
  termios.c_cflag &= ~PARENB;
  termios.c_cflag &= ~CSTOPB;
  termios.c_cflag &= ~CSIZE;
  termios.c_cflag |= CS8;
  
  // Set Speed into termios structure
  if (cfsetispeed(&termios, SER_Baudrate)) {
    logError(errno, "cfsetispeed");
    SER_close(SERfd);
    exit(-1);
  }
  if (cfsetospeed(&termios, SER_Baudrate)) {
    logError(errno, "cfsetospeed");
    SER_close(SERfd);
    exit(-1);
  }
  // Disable Echo, make Data=8, Stop=7, no parity, disable all controls
  // see "man cfmakeraw"
  cfmakeraw(&termios);
  tcflush(SERfd, TCIFLUSH);
  // set new attribute
  if (tcsetattr(SERfd, TCSANOW, &termios)) {
    logError(errno, "tcsetattr()");
    SER_close(SERfd);
    exit(-1);
  }
  SERrxlen = 0;
  return (SERfd);
}



int SER_close(int SERfd) {
  if (SERfd != 0) {
    SERrxlen = 0;
    tcsetattr(SERfd, TCSANOW, &OldTermIOS);
    return(close(SERfd));
  } else {
    return (0);
  }
}



int SER_SendBlock(int SERfd, unsigned char cmd, const char *buffer, unsigned short size) {
  int sendbytes = 0;
  unsigned short crc = 0;
  tRS232header head;
  head.h.id  = FRAMESTARTID;
  head.h.len = size+1;
  head.h.cmd = cmd;
  crcCCITTStream(&crc, head.data, sizeof(head));
  crcCCITTStream(&crc, buffer, (short) size);
  write(SERfd, head.data,  sizeof(head));
  while (sendbytes < size) {
    sendbytes += write(SERfd, buffer + sendbytes, (int)(size - sendbytes));
    if (sendbytes != size) {
      logError(0, "SER_SendBlock Buffer Overflow");
      if (tcdrain(SERfd)) {
        // Function not supported?
        logError(errno, "SER_SendBlock: tcdrain not supp.");
        exit(-1);
      }
    } //endif not full send - wait
  }
  crc = swap16(crc);
  write(SERfd, &crc, 2);
  return (0);
}



int SER_Receive(int SERfd, char *rxbuf, unsigned int maxsize) {
  int rdbytes, pktlen = 0;
  clock_t rx_time = clock();  // TimeOut-Behandlung
  if (((rx_time - last_rx_time) *100 / CLOCKS_PER_SEC) > SER_Timeout) {
    SERrxlen = 0;
  }
  last_rx_time = rx_time;
  // Auslesen bzw. empfangen von Bytes...
  if ((rdbytes = read(SERfd, &SERbuffer.data[SERrxlen], sizeof(SERbuffer) - SERrxlen)) > 0) {
    SERrxlen += rdbytes; // Anzahl gelesener Bytes hinzu.
    if ((SERrxlen > 0)&&(SERbuffer.h.id != FRAMESTARTID)) SERrxlen = 0;	// Kill ungültige
    if ((SERrxlen > 4) && (SERrxlen > (SERbuffer.h.len+4))) {  // Paket vollständig?
      pktlen = (int) SERbuffer.h.len + 5;
      if (crcCCITTBlock(SERbuffer.data, pktlen) == 0) {  // test CRC
        if (pktlen > maxsize) {
          pktlen = maxsize;
        }
	memcpy(rxbuf, SERbuffer.data, pktlen);
      } else {
        pktlen = -1; // CRC-Fehler
      } //if CRC-Check
      SERrxlen = 0;
    }   // Paket vorhanden
  } //if was RXed
  return (pktlen);
}


int SER_SendReceive(int SERfd, char Cmd, const char *SendBuf, unsigned short SendLen, char *ReceiveBuf, unsigned short MaxReceive) {
  int Sret, result, retries, timeout;
  struct timeval COMtimeout;  
  fd_set SERfds;      
  if (SendLen > SER_BUFSIZE-8)  return -EINVAL;
  if (MaxReceive > SER_BUFSIZE) return -EINVAL;
  SER_SendBlock(SERfd, Cmd, SendBuf, SendLen); 
  result  = -1;
  retries = SER_RETRIES;
  timeout = SER_PollDelay+SER_CALCTRANSFERDELAY(SendLen+MaxReceive);
  FD_ZERO(&SERfds);
  COMtimeout.tv_sec = 0;
  do {
    COMtimeout.tv_usec = timeout;		// Timeout neu setzen
    FD_SET(SERfd, &SERfds);
    Sret = select(SERfd+1, &SERfds, NULL, NULL, &COMtimeout);	// select()
    switch (Sret) {
      case -1: 
	return -1;
      case 0:	// Timeout
	if (--retries > 0) {
          SER_SendBlock(SERfd, Cmd, SendBuf, SendLen);	// Re-Send Block again
	} else return -1;
	break;
      default:
        result = SER_Receive(SERfd, ReceiveBuf, MaxReceive);
	break;
    } // hctiws Sret
  } while (result<=0);
  return result;
}
