/*
 * BIN -> DV-Firmware Utility
 * --------------------------
 *
 * Copyright (c) 2010 by Jan Alte, DO1FJN
 *
 *
 * bin2dvfw.c (main source file of dfu-loader)
 *
 * Author: Jan Alte, DO1FJN
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>

#include "checksums.h"
#include "fileIO.h"


#define PRG_NAME	"bin2dvfw"
#define VERSIONSTR	"Version 0.1\n\n"
#define DEFAULTNAME	"DV-Firmware"
#define FILESUFFIXSTR	"fw"

#define BOOTLOADERSIZE	0x4000L

#define DVDEVICE	0x1000
#define DVPRODUCTID	0x92AD
#define DVVENDORID	0x0403



typedef struct __attribute__((__packed__)) {
  unsigned short bcdDevice;
  unsigned short idProduct;	
  unsigned short idVendor;	//
  unsigned short bcdDFU;	//
  char		 DFUsignature[3];
  unsigned char  length;
  unsigned long  crc;		// 32bit CR
} tDFUfilesuffix;



int dfu_append_filesuffix(int FILEfd, tDFUfilesuffix *Suffix) {
  Suffix->bcdDFU = 0x0100;		// DFU-Filesuffix Version 1.00
  Suffix->DFUsignature[0] = 'U';
  Suffix->DFUsignature[1] = 'F';
  Suffix->DFUsignature[2] = 'D';
  Suffix->length = (unsigned char)sizeof(tDFUfilesuffix);
  crc32Stream(&Suffix->crc, (char *)Suffix, sizeof(tDFUfilesuffix)-4);
  if (FILEfd>2) {
    int len;
    lseek(FILEfd, 0L, SEEK_END);
    len = write(FILEfd, Suffix, sizeof(tDFUfilesuffix));
    return (len != sizeof(tDFUfilesuffix))?-1:0;
  }
  return 0;
}


int make_fwfile(const char *filename, const char *firmwarename) {
  int			INfile, OUTfile, retcode = -1;
  ssize_t		binsize, fwsize;
  char			*FwBuffer;
  tDFUfilesuffix	FwSuffix;
  
  if (file_exists(filename)) {
    INfile = open(filename, O_RDONLY);
    if (INfile < 0) {
      file_error(errno, "OPEN FILE");
      return -1;
    }
    binsize = lseek(INfile, 0L, SEEK_END);	// Get Size of Binary
    if (binsize > 0x4000000L) {	// only for fw smaller 64MByte!!!
      fprintf(stderr, "Warning: Selected file is to large!\n");
      close(INfile);
      return -1;      
    } else if (binsize <= BOOTLOADERSIZE) {
      fprintf(stderr, "Warning: Selected file is smaller then bootloader!\n");
      close(INfile);
      return -1;        
    }
    FwBuffer = (char *)malloc(binsize-BOOTLOADERSIZE);
    if (FwBuffer == NULL) {
      fprintf(stderr, "Error: Can't alloc file-buffer!\n");
      close(INfile);
      return -1;
    } // fi
    lseek(INfile, BOOTLOADERSIZE, SEEK_SET);
    fwsize = read(INfile, FwBuffer, binsize-BOOTLOADERSIZE);	// read hole firmware
    if (fwsize == (binsize-BOOTLOADERSIZE)) {
      // open temporary firmware file
      char *tempfilename = (char *)malloc(strlen(filename)+5);    
      FwSuffix.crc = 0xffffffff;	// init Checksum
      crc32Stream(&FwSuffix.crc, FwBuffer, fwsize);
      FwSuffix.bcdDevice = DVDEVICE;
      FwSuffix.idProduct = DVPRODUCTID;
      FwSuffix.idVendor  = DVVENDORID;
      if (tempfilename!=NULL) {
	sprintf(tempfilename, "%s.tmp", filename);
        OUTfile = open(tempfilename, O_WRONLY|O_CREAT|O_TRUNC|O_APPEND, 0644);
	if (OUTfile > 0) {
	  ssize_t outsize = write(OUTfile, FwBuffer, fwsize);
	  if ((dfu_append_filesuffix(OUTfile, &FwSuffix)!=0)||(outsize!=fwsize)) {
	    fprintf(stderr, "Error: Writing firmware fails!\n");
	    if (outsize < 0) file_error(errno, "WR TMP FILE");
	  } else // fi error
	    retcode = 0;
	  close(OUTfile);
	} // fi valid outfile
	if (retcode==0) {
          file_rename(tempfilename, firmwarename);
	} else {
	  file_delete(tempfilename);
	}
	free(tempfilename);
      } else {
	fprintf(stderr, "Error: Not enough mem for temporary filename\n");
      } // esle malloc err
    } else {
      if (fwsize < 0) {
         file_error(errno, filename);
       } else {
	 fprintf(stderr, "\nError: Can't read complete file!\n");	      
       } // esle      
    } // esle
    free(FwBuffer);
    close(INfile);
  } else {
    fprintf(stderr, "File '%s' doesn't exists.\n", filename);
  }
  return retcode;
}


void show_help(void) {
  printf("Usage: "PRG_NAME" [options] ...\n"
         "  -h --help\t\t\tPrint this help message\n"
         "  -V --version\t\t\tPrint the version number\n"
         "  -f --file <filename>\tSpecify binary fw-image"
         "  -n --name <name>\tSpecify filename of new firmware file\n\n");
}


const static char prg_options[] =
  "hVf:n:";

const static struct option prg_optslong[] = {
  { "help", 0, 0, 'h' },
  { "version", 0, 0, 'V' },
  { "file", 1, 0, 'f' },
  { "name", 1, 0, 'f' }
};


int main(int argc, char **argv) {
  int  exitcode  = 0, option;
  char *filename = NULL;
  char *fwname   = NULL;
  
  printf("\nDV-Firmware Make Utility - (C) 2010 by DO1FJN / digisolutions.de\n"
         "This program is Free Software and has ABSOLUTELY NO WARRANTY\n\n");
  do {
    int option_index = 0;
    option = getopt_long(argc, argv, prg_options, prg_optslong, &option_index);
    switch (option) {
      case 'h':
        show_help();
        exit(2);
        break;
      case 'V':
	printf(VERSIONSTR);
	break;
      case 'f':
	if (exitcode==0) filename = optarg;
	exitcode++;
        break;
      case 'n':
       fwname = (char *)malloc(strlen(optarg)+strlen(FILESUFFIXSTR)+2);
       if (fwname) sprintf(fwname, "%s.%s", optarg, FILESUFFIXSTR);
       break;
    } // hctiws
  } while (option != -1);
  
  if (fwname == NULL) {	// set Default name...
    fwname = (char *)malloc(strlen(DEFAULTNAME)+strlen(FILESUFFIXSTR)+2);
    if (fwname) sprintf(fwname, "%s.%s", DEFAULTNAME, FILESUFFIXSTR);
  }
  
  if (fwname != NULL) switch (exitcode) {
    case 0:
      exit(3);
    case 1:
      if (filename == NULL) {
	exitcode = -1;
	fprintf(stderr, "You need to specify a filename to -f\n");
	show_help();
      } else {
	exitcode = make_fwfile(filename, fwname);
      }
      break;
    default:
      fprintf(stderr, "Too many options!\n");
      show_help();
      exitcode = -2;
      break;
  } // hctiws
  if (fwname != NULL) 
    free(fwname);
  if (exitcode)
    printf("terminated!\n");
  else
    printf("done.\n");
  return exitcode;  
}

