/*
 * fileIO.c
 *
 * simple file functions and a logging mechanism
 *
 * Author: Jan Alte, DO1FJN
 *
 *
 * This file is part of the DFU Serial Loader (dfu-loader).
 * For general infomation about dfu-loader see "dfu_serial_main.c".
 *
 * dfu-loader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * dfu-loader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "fileIO.h"

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>		// Error-Nummer für logError
#include <sys/stat.h>


int loglevel = 0;


void logInit(const char *LogFilename) {
  freopen(LogFilename, "a", stderr);
  errno = 0;
}


void logError(int err, const char *onWhat){
  char timebuf[20];
  time_t now;
  if ((int)stderr > 2) {	// Write to File?
    now = time(NULL);		// Get the current time.
    strftime(&timebuf[0], sizeof(timebuf), "%Y-%m-%d %H:%M.%S", localtime(&now));
    fprintf(stderr, "%s: %s\n", timebuf, onWhat);	// save error message
  } else {
    fprintf(stderr, "%s\n", onWhat);			// display error message w/o time
  }
  if (err != 0) {
    fprintf(stderr, "-> %d: %s\n", err, strerror(err));	// detailed system err
//    errno = 0;
  }
  fflush(stderr);
  if (loglevel > 0) {
    printf("[%s]\n", onWhat);	// display error message  
    fflush(stdout);  
  }
} // logerror()



void file_error(int err, const char *pre_msg) {
  char errormsg[100]; 
  switch(err) {
    case EBADF: sprintf(errormsg, "%s: EBADF - bad filedescriptor", pre_msg); break;
    case EINTR: sprintf(errormsg, "%s: EINTR - interrupted", pre_msg); break;
    case EISDIR:
      sprintf(errormsg, "%s: EISDIR - this file is a directory.", pre_msg);
      break;    
    case EMFILE:
      sprintf(errormsg, "%s: EMFILE - too many filedescripors open.", pre_msg);
      break;
    case ENFILE:
      sprintf(errormsg, "%s: ENFILE - system has too many filedescripors open.", pre_msg);
      break;
    case ENOENT:
      sprintf(errormsg, "%s: ENOENT - the named file does not exist.", pre_msg);
      break;
    case ENOSPC:
      sprintf(errormsg, "%s: ENOSPC - no disk space left.", pre_msg);
      break;
    case ENXIO:
      sprintf(errormsg, "%s: ENXIO - file is a FIFO (can't opebn for reading).", pre_msg);
      break;
    case EROFS:      
      sprintf(errormsg, "%s: EROFS - read only file system.", pre_msg);
      break;
    case EEXIST:
      sprintf(errormsg, "%s: EEXIST - named file already exists.", pre_msg);
      break;      
    case EACCES:
      sprintf(errormsg, "%s: EACCES - no permission to do this.", pre_msg);
      break;
    case EFBIG:
      sprintf(errormsg, "%s: EFBIG - size too large.", pre_msg);
      break;
    case EIO:
      sprintf(errormsg, "%s: EIO - hardware error", pre_msg);
      break;
    case EPIPE:
      sprintf(errormsg, "%s: EPIPE - write to a closed pipe (FIFO) - nobody is reading it.", pre_msg);
      break;      
    default: 
      sprintf(errormsg, "%s: unknown error-code (%d).", pre_msg, errno); break;
  }
  logError(err, errormsg);
} //file_error



int file_exists(const char *fname) {
  int RetCode;
  struct stat fileinfo;
  RetCode = stat(fname, &fileinfo);
  return (RetCode==0)?1:0;
}


void file_rename(const char *fname_curr, const char *fname) {
  int RetCode;  
  if (file_exists(fname)) {	// Test if file fname exists...
    char *fname_backup = (char *)malloc(strlen(fname)+1);
    if (fname_backup != NULL) {
      sprintf(fname_backup, "%s~", fname);
      RetCode = rename(fname, fname_backup);
      if (RetCode != 0) {
        printf("Could not backup file (%s to %s).\n", fname_curr, fname);
        file_error(errno, "BACKUP");
      } // fi error
      free(fname_backup);
    } else {	// oh now: No mem 4 such a little thing...
      fprintf(stderr, "Could not alloc a piece of memory for the backup-filename!\n");
    }
  } // fi backup 
  RetCode = rename(fname_curr, fname);
  if (RetCode != 0) {
    printf("Could not rename file (%s to %s).\n", fname_curr, fname);
    file_error(errno, "RENAME");
  } // fi error
}


void file_delete(const char *fname) {
  int RetCode = remove(fname);
  if (RetCode != 0) {
    printf("Could not remove file (%s).\n", fname);
    file_error(errno, "REMOVE");
  } // fi error
}
